require "./ring.cr"
require "mutex"
require "openssl"

module Crystal::Scatter 

    class Daemon
        getter weight : UInt64
        getter location : Array(Int32)
        getter url : String
        def initialize(@weight : UInt64, @location : Array(Int32), @url : String)
        end
    end

    class MetaRing
        getter rings : Array(Ring)
        getter ring_graph : RingGraph

        def initialize(@shards : UInt32, @split_ratio : BigRational = BigRational.new(1,2))
            @rings = Array(Ring).new
            @ring_graph = RingGraph.new
            @lock = Mutex.new
        end

        private def add_impl(element : Daemon)
           @ring_graph.add(element.weight, element.location, element.url)
        end

        private def update_impl
            @ring_graph.generate_ring @split_ratio
            ring = Ring.new
            @ring_graph.snapshot do |daemon|
                ring.push Slice.new(daemon.range_effector.not_nil!.[0],daemon.range_effector.not_nil!.[1],daemon.reach_at)
            end
            @rings.push ring
        end

        private def hash_impl(data : String)
            ctx = OpenSSL::Digest.new("md4")
            ctx << data
            digest = ctx.digest
            tag = IO::ByteFormat::BigEndian.decode(UInt64, digest)
            tag ^= IO::ByteFormat::BigEndian.decode(UInt64, digest+8)
            return tag
        end

        private def get_slice_from_hash(ring : Ring, h : UInt64)
            value = ring.find do |slice|
                h>=slice.s_begin && slice.s_end>=h
            end
            return value.not_nil!
        end

        def add(element : Daemon)
            @lock.synchronize do
                add_impl element
                update_impl
            end
        end

        def add(elements : Array(Daemon))
            @lock.synchronize do
                elements.each do |element|
                    add_impl element
                end
                update_impl
            end
        end

        def hash(data : String)
            hash_impl data
        end

        def get_slices_for(data : String)
            ret = Array(Slice).new(@shards)
            ring? = nil : Ring | Nil
            @lock.synchronize do
                ring? = @rings.last
            end
            ring = ring?.not_nil!
            h = hash_impl data
            space=UInt64::MAX/@shards
            targets = Array(UInt64).new
            (1..@shards).each do
                targets << h
                h+=space
            end
            slices = Array(Slice).new
            targets.each do |h|
                slices << get_slice_from_hash(ring, h)
            end
            return slices
        end
    end

end